# from universi.exceptions import LintingError
# from universi.structure.common import Sentinel
# from universi.structure.schemas import (
#     AlterSchemaInstruction,
#     OldSchemaDidntHaveField,
#     OldSchemaFieldWas,
#     OldSchemaHadField,
# )
# from universi.structure.versions import AbstractVersionChange, Versions


# def check_old_schema_field_was(
#     change: OldSchemaFieldWas,
#     alter_schema_instruction: AlterSchemaInstruction,
# ):
#     actual_field_in_old_schema = alter_schema_instruction.old_model.__fields__[
#         change.field_name
#     ]
#     for attr_name, field in change.field_changes.__dict__.items():
#         if field is not Sentinel:
#             expected_old_value = getattr(
#                 actual_field_in_old_schema.field_info,
#                 attr_name,
#             )
#             if expected_old_value != field:
#                 raise LintingError(
#                     f"Expected value {expected_old_value} doesn't match the field value {field} for attribute {attr_name} in model {alter_schema_instruction.old_model}",
#                 )
#     if change.type is not Sentinel and (
#         alter_schema_instruction.old_model.__annotations__[change.field_name]
#         != change.type
#     ):
#         raise LintingError(
#             f"Field type mismatch for field {change.field_name} in model {alter_schema_instruction.old_model}",
#         )


# def check_old_schema_had_field(change: OldSchemaHadField, alter_schema_instruction):
#     if change.field_name not in alter_schema_instruction.old_model.__fields__:
#         raise LintingError(
#             f"Field {change.field_name} not found in old model {alter_schema_instruction.old_model}",
#         )
#     actual_field_in_old_schema = alter_schema_instruction.old_model.__fields__[
#         change.field_name
#     ]
#     for attr_name in actual_field_in_old_schema.field_info.__slots__:
#         expected_old_value = getattr(change.field, attr_name)
#         actual_old_value = getattr(actual_field_in_old_schema.field_info, attr_name)
#         if expected_old_value != actual_old_value:
#             raise LintingError(
#                 f"Attr {attr_name} is not equal in the migration ({expected_old_value}) and in actual old schema ({actual_old_value})",
#             )
#     if change.field_name in alter_schema_instruction.model.__fields__:
#         raise LintingError(
#             f"Unexpected field {change.field_name} found in new model {alter_schema_instruction.model}",
#         )


# def check_old_schema_didnt_have_field(
#     change: OldSchemaDidntHaveField,
#     alter_schema_instruction: AlterSchemaInstruction,
# ):
#     if change.field_name not in alter_schema_instruction.schema.__fields__:
#         raise LintingError(
#             f"Field {change.field_name} missing in new model {alter_schema_instruction.schema}",
#         )
#     if change.field_name in alter_schema_instruction.old_model.__fields__:
#         raise LintingError(
#             f"Unexpected field {change.field_name} found in old model {alter_schema_instruction.old_model}",
#         )


# def check_equivalent_fields(
#     alter_schema_instruction: AlterSchemaInstruction,
#     equivalent_fields,
# ):
#     for field_name in equivalent_fields:
#         old_model_field = alter_schema_instruction.old_model.__fields__[field_name]
#         new_model_field = alter_schema_instruction.schema.__fields__[field_name]

#         if old_model_field.type_ != new_model_field.type_:
#             raise LintingError(
#                 f"Field {field_name} is not equivalent by type in {alter_schema_instruction.old_model} and {alter_schema_instruction.schema}",
#             )

#         for attr_name in old_model_field.field_info.__slots__:
#             old_model_attr = getattr(old_model_field.field_info, attr_name)
#             new_model_attr = getattr(new_model_field.field_info, attr_name)
#             if old_model_attr != new_model_attr:
#                 raise LintingError(
#                     f"Field {field_name} is not equivalent by {attr_name} in {alter_schema_instruction.old_model} and {alter_schema_instruction.schema}",
#                 )


# def check_alter_schema_instruction(alter_schema_instruction: AlterSchemaInstruction):
#     for change in alter_schema_instruction.changes:
#         if isinstance(change, OldSchemaFieldWas):
#             check_old_schema_field_was(change, alter_schema_instruction)
#         elif isinstance(change, OldSchemaHadField):
#             check_old_schema_had_field(change, alter_schema_instruction)
#         elif isinstance(change, OldSchemaDidntHaveField):
#             check_old_schema_didnt_have_field(change, alter_schema_instruction)
#         else:
#             raise LintingError(
#                 f"Unexpected change of type {type(change)} in model {alter_schema_instruction.old_model}",
#             )

#     altered_fields = {change.field_name for change in alter_schema_instruction.changes}
#     equivalent_fields = (
#         set(alter_schema_instruction.old_model.__fields__) - altered_fields
#     )
#     fields_that_are_in_only_one_of_the_models = (
#         alter_schema_instruction.old_model.__fields__.keys()
#         ^ alter_schema_instruction.schema.__fields__.keys()
#     )

#     fields_that_are_not_altered_but_are_in_only_one_of_the_models = (
#         fields_that_are_in_only_one_of_the_models - altered_fields
#     )
#     if fields_that_are_not_altered_but_are_in_only_one_of_the_models:
#         raise LintingError(
#             f"Fields {fields_that_are_not_altered_but_are_in_only_one_of_the_models} are not altered but are in only one of the models: {alter_schema_instruction.old_model} and {alter_schema_instruction.schema}",
#         )

#     check_equivalent_fields(alter_schema_instruction, equivalent_fields)


# def check_version_change(version_change: type[AbstractVersionChange]):
#     for alter_schema_instruction in version_change.alter_schema_instructions:
#         check_alter_schema_instruction(alter_schema_instruction)


# def lint_all_schemas(versions: Versions):
#     for version in versions.versions:
#         for version_change in version.version_changes:
#             check_version_change(version_change)



import ast
import importlib
import inspect
import json
import os
import pprint
import shutil
import sys
from collections.abc import Sequence
from copy import deepcopy
from datetime import date
from enum import Enum, auto
from pathlib import Path
from types import GenericAlias, LambdaType, ModuleType
from typing import Any, TypeAlias, get_args, get_origin

from pydantic import BaseConfig, BaseModel
from pydantic.fields import ModelField
from typing_extensions import assert_never
from pydantic.fields import FieldInfo as PydanticFieldInfo

from ._utils import Sentinel
from .exceptions import CodeGenerationError, InvalidGenerationInstructionError
from .fields import FieldInfo
from universi.structure.enums import AlterEnumSubInstruction, EnumDidntHaveMembersInstruction, EnumHadMembersInstruction
from universi.structure.schemas import (
    AlterSchemaInstruction,
    OldSchemaDidntHaveField,
    OldSchemaFieldWas,
    OldSchemaHadField,
)
from universi.structure.versions import Version, Versions
from pydantic.typing import convert_generics

FieldNameT: TypeAlias = str
dict_of_empty_field_info = {k: getattr(PydanticFieldInfo(), k) for k in PydanticFieldInfo.__slots__}


# TODO: Add enum alteration here
def regenerate_dir_to_all_versions(template_module: ModuleType, versions: Versions):
    schemas = {k: (v, _get_fields_for_model(v)) for k, v in deepcopy(versions.versioned_schemas).items()}
    enums = {k: (v, {member.name: member.value for member in v}) for k, v in deepcopy(versions.versioned_enums).items()}

    for version in versions.versions:
        # NOTE: You'll have to use relative imports

        _generate_versioned_directory(template_module, schemas, enums, version.date)
        _apply_migrations(version, schemas, enums)

    base_package = template_module.__name__.split(".")[0]
    for module_name, module_value in sys.modules.copy().items():
        if module_name.startswith(base_package) and module_name in sys.modules:
            importlib.reload(module_value)


def _apply_migrations(
    version: Version,
    schemas: dict[
        str,
        tuple[type[BaseModel], dict[FieldNameT, tuple[type[BaseModel], ModelField]]],
    ],
    enums: dict[str, tuple[type[Enum], dict[str, Any]]],
):
    for version_change in version.version_changes:
        _apply_alter_schema_instructions(schemas, version_change.alter_schema_instructions)
        _apply_alter_enum_instructions(enums, version_change.alter_enum_instructions)


def _apply_alter_schema_instructions(
    schemas: dict[str, tuple[type[BaseModel], dict[FieldNameT, tuple[type[BaseModel], ModelField]]]],
    alter_schema_instructions: Sequence[AlterSchemaInstruction],
):
    for alter_schema_instruction in alter_schema_instructions:
        schema = alter_schema_instruction.schema
        schema_path = schema.__module__ + schema.__name__
        schema_field_info_bundle = schemas[schema_path]
        field_name_to_field_model = schema_field_info_bundle[1]
        for field_change in alter_schema_instruction.changes:
            if isinstance(field_change, OldSchemaDidntHaveField):
                # TODO: Check that the user doesn't pop it and change it at the same time
                # TODO: Add a check that field actually exists (it's very necessary!)
                field_name_to_field_model.pop(field_change.field_name)

            elif isinstance(field_change, OldSchemaFieldWas):
                # TODO: Add a check that field actually exists (it's very necessary!)
                model_field = field_name_to_field_model[field_change.field_name][1]
                if field_change.type is not Sentinel:
                    if model_field.annotation == field_change.type:
                        raise InvalidGenerationInstructionError(
                            f"You tried to change the type of field '{field_change.field_name}' to '{field_change.type}' in {schema.__name__} but it already has type '{model_field.annotation}'"
                        )
                    setattr(model_field, "annotation", field_change.type)
                    setattr(model_field, "type_", convert_generics(field_change.type))
                    setattr(model_field, "outer_type_", field_change.type)
                field_info = model_field.field_info

                if not isinstance(field_info, FieldInfo):
                    dict_of_field_info = {k: getattr(field_info, k) for k in field_info.__slots__}
                    if dict_of_field_info == dict_of_empty_field_info:
                        field_info = FieldInfo()
                        model_field.field_info = field_info
                    else:
                        raise InvalidGenerationInstructionError(
                            f"You have defined a Field using pydantic.fields.Field but you must use universi.Field in {schema.__name__}"
                        )
                for attr_name in field_change.field_changes.__dataclass_fields__:
                    attr_value = getattr(field_change.field_changes, attr_name)
                    if attr_value is not Sentinel:
                        setattr(field_info, attr_name, attr_value)
                        field_info._universi_field_names.add(attr_name)
            elif isinstance(field_change, OldSchemaHadField):
                field_name_to_field_model[field_change.field_name] = (
                    schema,
                    ModelField(
                        name=field_change.field_name,
                        type_=field_change.type,
                        field_info=field_change.field,
                        class_validators=None,
                        model_config=BaseConfig,
                    ),
                )
            else:
                assert_never(field_change)


def _apply_alter_enum_instructions(
    enums: dict[str, tuple[type[Enum], dict[str, Any]]],
    alter_enum_instructions: Sequence[AlterEnumSubInstruction],
):
    for alter_enum_instruction in alter_enum_instructions:
        enum = alter_enum_instruction.enum
        enum_path = enum.__module__ + enum.__name__
        enum_member_to_value = enums[enum_path]
        if isinstance(alter_enum_instruction, EnumDidntHaveMembersInstruction):
            for member in alter_enum_instruction.members:
                if member not in enum_member_to_value[1]:
                    raise InvalidGenerationInstructionError(
                        f"Enum member '{member}' was not found in enum '{enum_path}'"
                    )
                enum_member_to_value[1].pop(member)
        elif isinstance(alter_enum_instruction, EnumHadMembersInstruction):
            for member, member_value in alter_enum_instruction.members.items():
                if member in enum_member_to_value[1] and enum_member_to_value[1][member] == member_value:
                    raise InvalidGenerationInstructionError(
                        f"Enum member '{member}' already exists in enum '{enum_path}' with the same value",
                    )
                else:
                    enum_member_to_value[1][member] = member_value
        else:
            assert_never(alter_enum_instruction)


def _get_versioned_schema_dir_name(template_module: ModuleType, version: date) -> Path:
    template_dir = _get_package_path_from_module(template_module)

    version_as_str = version.isoformat().replace("-", "_")
    return template_dir.with_name("v" + version_as_str)


def _get_package_path_from_module(template_module: ModuleType) -> Path:
    file = inspect.getsourcefile(template_module)

    # I am too lazy to reproduce this error correctly
    if file is None:  # pragma: no cover
        raise CodeGenerationError(f"Module {template_module} has no source file")
    file = Path(file)
    if not file.name == "__init__.py":
        raise CodeGenerationError(f"Module {template_module} is not a package")
    return file.parent


def _generate_versioned_directory(
    template_module: ModuleType,
    schemas: dict[
        str,
        tuple[type[BaseModel], dict[FieldNameT, tuple[type[BaseModel], ModelField]]],
    ],
    enums: dict[str, tuple[type[Enum], dict[str, Any]]],
    version: date,
):
    assert template_module.__file__ is not None
    dir = _get_package_path_from_module(template_module)
    version_dir = _get_versioned_schema_dir_name(template_module, version)
    version_dir.mkdir(exist_ok=True)
    # [universi, structure, schemas]
    template_module_python_path_parts = template_module.__name__.split(".")
    # [home, foo, bar, universi, structure, schemas]
    template_module_path_parts = Path(template_module.__file__).parent.parts
    # [home, foo, bar] = [home, foo, bar, universi, structure, schemas][:-3]
    root_module_path = Path(*template_module_path_parts[: -len(template_module_python_path_parts)])
    for subroot, dirnames, filenames in os.walk(dir):
        original_subroot = Path(subroot)
        versioned_subroot = version_dir / original_subroot.relative_to(dir)
        if "__pycache__" in dirnames:
            dirnames.remove("__pycache__")
        for dirname in dirnames:
            (versioned_subroot / dirname).mkdir(exist_ok=True)
        for filename in filenames:
            original_file = (original_subroot / filename).absolute()
            versioned_file = (versioned_subroot / filename).absolute()
            print(original_file)

            if filename.endswith(".py"):
                module_path = ".".join(
                    original_file.relative_to(root_module_path).with_suffix("").parts,
                )
                module = importlib.import_module(module_path)
                new_module_text = _modify_module(module, schemas, enums)
                versioned_file.write_text(new_module_text)
            else:
                shutil.copyfile(original_file, versioned_file)


def _get_fields_for_model(model: type[BaseModel]) -> dict[FieldNameT, tuple[type[BaseModel], ModelField]]:
    actual_fields: dict[FieldNameT, tuple[type[BaseModel], ModelField]] = {}
    for cls in model.__mro__:
        if cls is BaseModel:
            return actual_fields
        if not issubclass(cls, BaseModel):
            continue
        for field_name, field in cls.__fields__.items():
            if field_name not in actual_fields and field_name in cls.__annotations__:
                actual_fields[field_name] = (cls, field)
    else:
        raise CodeGenerationError(f"Model {model} is not a subclass of BaseModel")


def _modify_module(
    module,
    modified_schemas: dict[
        str,
        tuple[type[BaseModel], dict[str, tuple[type[BaseModel], ModelField]]],
    ],
    modified_enums: dict[str, tuple[type[Enum], dict[str, Any]]],
) -> str:
    try:
        parsed_file = ast.parse(inspect.getsource(module))
    except OSError as e:
        path = Path(module.__file__)
        if path.is_file() and path.read_text() == "":
            return ""
        # Not sure how to get here so this is just a precaution
        raise CodeGenerationError("Failed to get source code for module") from e  # pragma: no cover

    if module.__name__.endswith(".__init__"):
        module_name = module.__name__.removesuffix(".__init__")
    else:
        module_name = module.__name__
    body = ast.Module(
        [ast.ImportFrom(module="universi", names=[ast.alias(name="Field")], level=0)]
        + [
            _modify_cls(n, module_name, modified_schemas, modified_enums) if isinstance(n, ast.ClassDef) else n
            for n in parsed_file.body
        ],
        [],
    )

    return ast.unparse(body)


def _modify_cls(
    cls_node: ast.ClassDef,
    module_python_path: str,
    modified_schemas: dict[
        str,
        tuple[type[BaseModel], dict[str, tuple[type[BaseModel], ModelField]]],
    ],
    modified_enums: dict[str, tuple[type[Enum], dict[str, Any]]],
) -> ast.ClassDef:
    cls_python_path = module_python_path + cls_node.name
    if cls_python_path in modified_schemas:
        cls_node = _modify_schema_cls(cls_node, modified_schemas[cls_python_path][1])
    if cls_python_path in modified_enums:
        cls_node = _modify_enum_cls(cls_node, modified_enums[cls_python_path][1])

    if not cls_node.body:
        cls_node.body = [ast.Pass()]
    return cls_node


def _modify_schema_cls(
    cls_node: ast.ClassDef,
    actual_fields: dict[str, tuple[type[BaseModel], ModelField]],
) -> ast.ClassDef:
    body = [
        ast.AnnAssign(
            target=ast.Name(id=name, ctx=ast.Store()),
            annotation=ast.Name(id=custom_repr(field[1].annotation), ctx=ast.Load()),
            value=ast.Call(
                func=ast.Name(id="Field", ctx=ast.Load()),
                args=[],
                keywords=[
                    ast.keyword(
                        arg=attr,
                        value=ast.Name(
                            id=custom_repr(getattr(field[1].field_info, attr)),
                        ),
                    )
                    # TODO: We should lint the code to make sure that the user is not using pydantic.fields.Field instead of universi.Field
                    for attr in getattr(field[1].field_info, "_universi_field_names", ())
                ],
            ),
            simple=1,
        )
        for name, field in actual_fields.items()
    ]
    old_body = [n for n in cls_node.body if not isinstance(n, ast.AnnAssign | ast.Pass | ast.Ellipsis)]
    docstring = _pop_docstring_from_cls_body(old_body)
    cls_node.body = docstring + body + old_body

    return cls_node


def _modify_enum_cls(cls_node: ast.ClassDef, enum_info: dict[str, Any]) -> ast.ClassDef:
    new_body = [
        ast.Assign(
            targets=[ast.Name(id=member, ctx=ast.Store())],
            value=ast.Name(id=custom_repr(member_value)),
            lineno=0,
        )
        for member, member_value in enum_info.items()
    ]

    old_body = [n for n in cls_node.body if not isinstance(n, ast.AnnAssign | ast.Assign | ast.Pass | ast.Ellipsis)]
    docstring = _pop_docstring_from_cls_body(old_body)

    cls_node.body = docstring + new_body + old_body
    return cls_node


def _pop_docstring_from_cls_body(old_body: list[ast.stmt]) -> list[ast.stmt]:
    if (
        len(old_body) > 0
        and isinstance(old_body[0], ast.Expr)
        and isinstance(old_body[0].value, ast.Constant)
        and isinstance(old_body[0].value.value, str)
    ):
        return [old_body.pop(0)]
    else:
        return []


# The following is based on by Samuel Colvin's devtools


# TODO: Fully cover this with tests
def custom_repr(value: Any) -> Any:
    if isinstance(value, list | tuple | set | frozenset):
        return PlainRepr(value.__class__(map(custom_repr, value)))
    if isinstance(value, dict):
        return PlainRepr(value.__class__((custom_repr(k), custom_repr(v)) for k, v in value.items()))
    if isinstance(value, GenericAlias):
        return f"{custom_repr(get_origin(value))}[{', '.join(custom_repr(a) for a in get_args(value))}]"
    if isinstance(value, type):
        return value.__name__
    if isinstance(value, Enum):
        return PlainRepr(f"{value.__class__.__name__}.{value.name}")
    if isinstance(value, auto):
        return PlainRepr("auto()")
    if isinstance(value, LambdaType):
        return _find_a_lambda(inspect.getsource(value).strip())
    if inspect.isfunction(value):
        return PlainRepr(value.__name__)
    else:
        return PlainRepr(repr(value))


class PlainRepr(str):
    """
    String class where repr doesn't include quotes.
    """

    def __repr__(self) -> str:
        return str(self)


def _find_a_lambda(source: str) -> str:
    found_lambdas: list[ast.Lambda] = []
    for node in ast.walk(ast.parse(source)):
        if isinstance(node, ast.keyword) and node.arg == "default_factory" and isinstance(node.value, ast.Lambda):
            found_lambdas.append(node.value)
    if len(found_lambdas) == 1:
        return ast.unparse(found_lambdas[0])
    # These two errors are really hard to cover. Not sure if even possible, honestly :)
    elif len(found_lambdas) == 0:  # pragma: no cover
        raise InvalidGenerationInstructionError(
            f"No lambda found in default_factory even though one was passed: {source}"
        )
    else:  # pragma: no cover
        raise InvalidGenerationInstructionError("More than one lambda found in default_factory. This is not supported.")



import importlib.metadata

from .codegen import regenerate_dir_to_all_versions
from .fields import Field
from .header import api_version_var
from .routing import VersionedAPIRouter
from .structure import Versions

__version__ = importlib.metadata.version("universi")
__all__ = ["Field", "VersionedAPIRouter", "api_version_var", "regenerate_dir_to_all_versions", "Versions"]



import importlib
import inspect
import sys
from pathlib import Path
from types import ModuleType
from typing import Any

Sentinel: Any = object()


def get_another_version_of_cls(cls_from_old_version: type[Any], new_version_dir: Path) -> None:
    # version_dir = /home/ovsyanka/package/companies/v2021_01_01

    module_from_old_version = sys.modules[cls_from_old_version.__module__]
    module = get_another_version_of_module(module_from_old_version, new_version_dir)
    return getattr(module, cls_from_old_version.__name__)


# TODO: WHat if the user puts model in __init__.py?
def get_another_version_of_module(module_from_old_version: ModuleType, new_version_dir: Path):
    file = inspect.getsourcefile(module_from_old_version)
    if file is None:
        raise Exception(f"Model {module_from_old_version} is not defined in a file")

    # /home/ovsyanka/package/companies/latest/__init__.py
    file = Path(file)
    if file.name == "__init__.py":
        # /home/ovsyanka/package/companies/latest/
        file = file.parent
    # /home/ovsyanka/package/companies
    root_dir = new_version_dir.parent
    # latest/schemas
    relative_file = file.relative_to(root_dir).with_suffix("")
    # ['latest', 'schemas']
    relative_file_parts = relative_file.parts
    # package.companies.latest.schemas.Payable
    model_python_path = module_from_old_version.__name__
    # ['package', 'companies', 'latest', 'schemas']
    model_split_python_path = model_python_path.split(".")

    # ['package', 'companies', 'latest', 'schemas']
    #                           ^^^^^^
    #                           index = -2

    # ['package', 'companies', 'v2021_01_01', 'schemas']
    model_split_python_path[-len(relative_file_parts)] = new_version_dir.name
    # package.companies.v2021_01_01.schemas
    new_model_module_python_path = ".".join(model_split_python_path)
    module = importlib.import_module(new_model_module_python_path)
    return module



import functools
import inspect
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any, ParamSpec, TypeVar

from .common import Endpoint

_P = ParamSpec("_P")
_T = TypeVar("_T")


@dataclass
class AlterResponseInstruction:
    endpoints: tuple[Endpoint, ...]
    method: Callable[[type, dict[str, Any]], None]
    owner: type = field(init=False)

    def __post_init__(self):
        assert inspect.signature(self.method).return_annotation is None
        assert (
            len(inspect.signature(self.method).parameters) == 2
        ), f"Method {self.method.__name__} must have 2 parameters: cls and data"
        annotation = inspect.signature(self.method).parameters["data"].annotation
        assert annotation == dict[str, Any], annotation

        functools.update_wrapper(self, self.method)

    def __set_name__(self, owner, name):
        self.owner = owner

    def __call__(self, data: dict[str, Any]) -> None:
        return self.method(self.owner, data)


def convert_response_to_previous_version_for(
    first_endpoint: Endpoint, /, *other_endpoints: Endpoint,
) -> "type[classmethod[Any, _P, None]]":
    def decorator(method: Callable[[object, dict[str, Any]], None]) -> Any:
        return AlterResponseInstruction((first_endpoint, *other_endpoints), method)

    return decorator  # pyright: ignore[reportGeneralTypeIssues]



import datetime
from collections.abc import Callable
from copy import deepcopy
from enum import Enum
import functools
import inspect
from pathlib import Path
from types import GenericAlias, MappingProxyType, ModuleType
from starlette.routing import request_response
from typing import (
    _SpecialForm,
    Annotated,
    Any,
    Collection,
    ParamSpec,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)
from typing_extensions import assert_never
import fastapi.routing
from fastapi.dependencies.utils import get_dependant, get_body_field, get_parameterless_sub_dependant
from fastapi.routing import APIRoute
from pydantic import BaseModel
from starlette._utils import is_async_callable
from starlette.routing import (
    BaseRoute,
)
from starlette.routing import Mount as Mount  # noqa
from typing_extensions import Self

from universi._utils import Sentinel, get_another_version_of_cls
from universi.codegen import _get_versioned_schema_dir_name
from universi.exceptions import RouterGenerationError, UniversiError
from universi.structure.common import Endpoint
from universi.structure.endpoints import (
    EndpointDidntExistInstruction,
    EndpointExistedInstruction,
    EndpointHadInstruction,
)
from universi.structure.versions import Versions
from fastapi.params import Depends
from threading import Lock
from typing import _BaseGenericAlias  # pyright: ignore

_T = TypeVar("_T", bound=Callable[..., Any])
annotation_resolution_lock = Lock()


def same_definition_as_in(t: _T) -> Callable[[Callable], _T]:
    def decorator(f: Callable) -> _T:
        return f  # pyright: ignore

    return decorator


def lock_cache(f: _T) -> _T:
    @functools.wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            with annotation_resolution_lock:
                return f(*args, **kwargs)
        finally:
            _memoized_change_versions_of_all_annotations.cache_clear()

    return cast(_T, wrapper)


class VersionedAPIRouter(fastapi.routing.APIRouter):
    @same_definition_as_in(fastapi.routing.APIRouter.__init__)
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._deleted_routes = []

    def only_exists_in_older_versions(self, endpoint: _T) -> _T:
        index = _get_route_index(self.routes, endpoint)
        if index is None:
            raise LookupError(f"Route not found on endpoint: '{endpoint.__name__}'")
        self._deleted_routes.append(self.routes.pop(index))
        return endpoint

    @lock_cache
    def create_versioned_copies(
        self,
        versions: Versions,
        *,
        latest_schemas_module: ModuleType | None,
    ) -> dict[datetime.date, Self]:
        router = self
        for route in router.routes + router._deleted_routes:
            if isinstance(route, APIRoute):
                if not is_async_callable(route.endpoint):
                    raise TypeError(
                        "All versioned endpoints must be asynchronous.",
                    )
                route.endpoint = versions.versioned(route.endpoint)
        routers = {}
        for version in versions.versions:
            if latest_schemas_module:
                version_dir = _get_versioned_schema_dir_name(latest_schemas_module, version.date)
                if not version_dir.is_dir():
                    raise RouterGenerationError(f"Versioned schema directory '{version_dir}' does not exist.")
                for route in router.routes:
                    if isinstance(route, APIRoute):
                        if route.response_model is not None:
                            route.response_model = _change_versions_of_all_annotations(
                                route.response_model, version_dir
                            )
                        route.endpoint = _change_versions_of_all_annotations(route.endpoint, version_dir)
                        route.dependant = get_dependant(path=route.path_format, call=route.endpoint)
                        route.body_field = get_body_field(dependant=route.dependant, name=route.unique_id)
                        for depends in route.dependencies[::-1]:
                            route.dependant.dependencies.insert(
                                0,
                                get_parameterless_sub_dependant(depends=depends, path=route.path_format),
                            )
                        route.app = request_response(route.get_route_handler())
            routers[version.date] = router
            router = deepcopy(router)
            for version_change in version.version_changes:
                for instruction in version_change.alter_endpoint_instructions:
                    original_route_index = _get_route_index(
                        router.routes,
                        instruction.endpoint,
                    )
                    if isinstance(instruction, EndpointDidntExistInstruction):
                        if original_route_index is None:
                            raise RouterGenerationError(
                                f"Endpoint '{instruction.endpoint.__name__}' you tried to delete in '{version_change.__name__}' doesn't exist in new version",
                            )
                        router.routes.pop(original_route_index)
                    elif isinstance(instruction, EndpointExistedInstruction):
                        if original_route_index is not None:
                            raise RouterGenerationError(
                                f"Endpoint '{instruction.endpoint.__name__}' you tried to re-create in '{version_change.__name__}' already existed in newer versions"
                            )
                        deleted_route_index = _get_route_index(
                            router._deleted_routes,
                            instruction.endpoint,
                        )
                        if deleted_route_index is None:
                            raise RouterGenerationError(
                                f"Endpoint '{instruction.endpoint.__name__}' you tried to re-create in '{version_change.__name__}' wasn't among the deleted routes"
                            )
                        router.routes.append(
                            router._deleted_routes.pop(deleted_route_index),
                        )
                    elif isinstance(instruction, EndpointHadInstruction):
                        if original_route_index is None:
                            raise RouterGenerationError(
                                f"Endpoint '{instruction.endpoint.__name__}' you tried to delete in '{version_change.__name__}' doesn't exist in new version"
                            )
                        route = router.routes[original_route_index]
                        for attr_name in instruction.attributes.__dataclass_fields__:
                            attr = getattr(instruction.attributes, attr_name)
                            if attr is not Sentinel:
                                assert getattr(route, attr_name) != attr
                                setattr(route, attr_name, attr)
                    else:
                        assert_never(instruction)
        return routers


def _change_versions_of_all_annotations(annotation: Any, version_dir: Path) -> Any:
    if isinstance(annotation, dict):
        return {
            _change_versions_of_all_annotations(key, version_dir): _change_versions_of_all_annotations(
                value, version_dir
            )
            for key, value in annotation.items()
        }

    elif isinstance(annotation, (list, tuple)):
        return type(annotation)(_change_versions_of_all_annotations(v, version_dir) for v in annotation)
    else:
        return _memoized_change_versions_of_all_annotations(annotation, version_dir)


# This cache is not here for speeding things up. It's for preventing the creation of copies of the same object
# because such copies could produce weird behaviors at runtime, especially if you/fastapi do any comparisons.
@functools.cache
def _memoized_change_versions_of_all_annotations(annotation: Any, version_dir: Path) -> Any:
    if isinstance(annotation, (_BaseGenericAlias, GenericAlias)):
        return _change_versions_of_all_annotations(get_origin(annotation), version_dir)[
            tuple(_change_versions_of_all_annotations(arg, version_dir) for arg in get_args(annotation))
        ]
    elif isinstance(annotation, Depends):
        return Depends(
            _change_versions_of_all_annotations(annotation.dependency, version_dir), use_cache=annotation.use_cache
        )
    elif isinstance(annotation, type):
        if issubclass(annotation, (BaseModel, Enum)):
            return get_another_version_of_cls(annotation, version_dir)
        else:
            return annotation
    # TODO: We need memoization for callables. Otherwise we're gonna get in trouble
    elif getattr(annotation, "__call__", None) is not None:
        if inspect.iscoroutinefunction(annotation):

            @functools.wraps(annotation)
            async def new_callable(*args: Any, **kwargs: Any) -> Any:
                return await annotation(*args, **kwargs)

        else:

            @functools.wraps(annotation)
            def new_callable(*args: Any, **kwargs: Any) -> Any:
                return annotation(*args, **kwargs)

        del new_callable.__wrapped__
        old_params = inspect.signature(annotation).parameters
        new_callable.__annotations__ = _change_versions_of_all_annotations(new_callable.__annotations__, version_dir)
        new_callable.__defaults__ = _change_versions_of_all_annotations(
            tuple(p.default for p in old_params.values() if p.default is not inspect.Signature.empty),
            version_dir=version_dir,
        )
        new_callable.__signature__ = _generate_signature(new_callable, old_params)
        return new_callable
    else:
        return annotation


def _generate_signature(new_callable: Callable, old_params: MappingProxyType[str, inspect.Parameter]):
    parameters = []
    default_counter = 0
    for param in old_params.values():
        if param.default is not inspect.Signature.empty:
            default = new_callable.__defaults__[default_counter]
            default_counter += 1
        else:
            default = inspect.Signature.empty
        parameters.append(
            inspect.Parameter(
                param.name,
                param.kind,
                default=default,
                annotation=new_callable.__annotations__.get(param.name, inspect.Signature.empty),
            )
        )
    return inspect.Signature(
        parameters=parameters,
        return_annotation=new_callable.__annotations__.get("return", inspect.Signature.empty),
    )


def _get_route_index(routes: list[BaseRoute], endpoint: Endpoint):
    for index, route in enumerate(routes):
        if isinstance(route, APIRoute) and (
            route.endpoint == endpoint or getattr(route.endpoint, "func", None) == endpoint
        ):
            return index
    return None



from contextvars import ContextVar
import datetime
import functools
from collections.abc import Callable, Sequence
from enum import Enum
from typing import Any, ClassVar, ParamSpec, TypeAlias, TypeVar, overload

from fastapi.routing import _prepare_response_content
from universi.exceptions import UniversiStructureError

from universi.header import api_version_var
from universi.structure.endpoints import AlterEndpointSubInstruction
from universi.structure.enums import AlterEnumSubInstruction

from .._utils import Sentinel
from .common import Endpoint, VersionedModel
from .responses import AlterResponseInstruction
from .schemas import AlterSchemaInstruction

_P = ParamSpec("_P")
_R = TypeVar("_R")
VersionDate: TypeAlias = datetime.date


class AbstractVersionChange:
    side_effects: ClassVar[bool] = False
    description: ClassVar[str] = Sentinel
    instructions_to_migrate_to_previous_version: ClassVar[
        Sequence[AlterSchemaInstruction | AlterEndpointSubInstruction | AlterEnumSubInstruction]
    ] = Sentinel
    alter_schema_instructions: ClassVar[Sequence[AlterSchemaInstruction]] = Sentinel
    alter_enum_instructions: ClassVar[Sequence[AlterEnumSubInstruction]] = Sentinel
    alter_endpoint_instructions: ClassVar[Sequence[AlterEndpointSubInstruction]] = Sentinel
    alter_response_instructions: ClassVar[dict[Endpoint, AlterResponseInstruction]] = Sentinel

    def __init_subclass__(cls) -> None:
        assert isinstance(cls.side_effects, bool)
        if cls.description is Sentinel:
            raise UniversiStructureError(f"Version change description is not set on '{cls.__name__}' but is required.")
        if cls.instructions_to_migrate_to_previous_version is Sentinel:
            raise UniversiStructureError(
                f"Attribute 'instructions_to_migrate_to_previous_version' is not set on '{cls.__name__}' but is required.",
            )
        if not isinstance(cls.instructions_to_migrate_to_previous_version, Sequence):
            raise UniversiStructureError(
                f"Attribute 'instructions_to_migrate_to_previous_version' must be a sequence in '{cls.__name__}'.",
            )
        for attr_name, attr_value in cls.__dict__.items():
            if not isinstance(attr_value, AlterResponseInstruction) and attr_name not in {
                "description",
                "side_effects",
                "instructions_to_migrate_to_previous_version",
                "__module__",
                "__doc__",
            }:
                raise UniversiStructureError(
                    f"Found: '{attr_name}' attribute in {cls.__name__}. Only migration instructions are allowed in version change classes.",
                )

        cls.alter_schema_instructions = []
        cls.alter_enum_instructions = []
        cls.alter_endpoint_instructions = []
        for alter_instruction in cls.instructions_to_migrate_to_previous_version:
            if isinstance(alter_instruction, AlterSchemaInstruction):
                cls.alter_schema_instructions.append(alter_instruction)
            elif isinstance(alter_instruction, AlterEnumSubInstruction):
                cls.alter_enum_instructions.append(alter_instruction)
            else:
                cls.alter_endpoint_instructions.append(alter_instruction)
        cls.alter_response_instructions = {
            endpoint: instruction
            for instruction in cls.__dict__.values()
            if isinstance(instruction, AlterResponseInstruction)
            for endpoint in instruction.endpoints
        }
        repetitions = set()
        for alter_instruction in cls.alter_schema_instructions:
            assert (
                alter_instruction.schema not in repetitions
            ), f"Model {alter_instruction.schema} got repeated. Please, merge these instructions."
            repetitions.add(alter_instruction.schema)

        if cls.mro() != [cls, AbstractVersionChange, object]:
            raise TypeError(
                f"Can't subclass {cls.__name__} as it was never meant to be subclassed.",
            )

    def __init__(self) -> None:
        raise TypeError(
            f"Can't instantiate {self.__class__.__name__} as it was never meant to be instantiated.",
        )


class Version:
    def __init__(
        self,
        date: VersionDate,
        *version_changes: type[AbstractVersionChange],
    ) -> None:
        self.date = date
        self.version_changes = version_changes


class Versions:
    def __init__(self, *versions: Version, api_version_var: ContextVar[VersionDate | None] = api_version_var) -> None:
        self.versions = versions
        self.api_version_var = api_version_var
        if sorted(versions, key=lambda v: v.date, reverse=True) != list(versions):
            raise ValueError(
                "Versions are not sorted correctly. Please sort them in descending order.",
            )

    @functools.cached_property
    def versioned_schemas(self) -> dict[str, type[VersionedModel]]:
        return {
            instruction.schema.__module__ + instruction.schema.__name__: instruction.schema
            for version in self.versions
            for version_change in version.version_changes
            for instruction in version_change.alter_schema_instructions
        }

    @functools.cached_property
    def versioned_enums(self) -> dict[str, type[Enum]]:
        return {
            instruction.enum.__module__ + instruction.enum.__name__: instruction.enum
            for version in self.versions
            for version_change in version.version_changes
            for instruction in version_change.alter_enum_instructions
        }

    @functools.cached_property
    def _version_changes_to_version_mapping(self) -> dict[type[AbstractVersionChange], VersionDate]:
        return {version_change: version.date for version in self.versions for version_change in version.version_changes}

    def is_active(self, version_change: type[AbstractVersionChange]) -> bool:
        api_version = self.api_version_var.get()
        if api_version is None:
            return False
        return self._version_changes_to_version_mapping[version_change] <= api_version

    # TODO: It might need caching for iteration to speed it up
    def data_to_version(
        self,
        endpoint: Endpoint,
        data: dict[str, Any],
        version: VersionDate,
    ) -> dict[str, Any]:
        for v in self.versions:
            if v.date <= version:
                break
            for version_change in v.version_changes:
                if endpoint in version_change.alter_response_instructions:
                    version_change.alter_response_instructions[endpoint](data)
        return data

    @overload
    def versioned(self, endpoint: Endpoint[_P, _R]) -> Endpoint[_P, _R]:
        ...

    @overload
    def versioned(
        self,
        endpoint: None = None,
    ) -> Callable[[Endpoint[_P, _R]], Endpoint[_P, _R]]:
        ...

    def versioned(
        self,
        endpoint: Endpoint | None = None,
    ) -> Callable[[Endpoint[_P, _R]], Endpoint[_P, _R]] | Endpoint[_P, _R]:
        if endpoint is not None:

            @functools.wraps(endpoint)
            async def decorator(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                return await self._convert_endpoint_response_to_version(
                    endpoint,
                    args,
                    kwargs,
                )

            decorator.func = endpoint  # pyright: ignore[reportGeneralTypeIssues]
            return decorator

        def wrapper(endpoint: Endpoint[_P, _R]) -> Endpoint[_P, _R]:
            @functools.wraps(endpoint)
            async def decorator(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                return await self._convert_endpoint_response_to_version(
                    endpoint,
                    args,
                    kwargs,
                )

            decorator.func = endpoint  # pyright: ignore[reportGeneralTypeIssues]
            return decorator

        return wrapper

    async def _convert_endpoint_response_to_version(
        self,
        endpoint: Endpoint,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        response = await endpoint(*args, **kwargs)
        api_version = self.api_version_var.get()
        if api_version is None:
            return response
        # TODO We probably need to call this in the same way as in fastapi instead of hardcoding exclude_unset.
        # We have such an ability if we force passing the route into this wrapper. Or maybe not... Important!
        response = _prepare_response_content(response, exclude_unset=False)
        return self.data_to_version(endpoint, response, api_version)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['universi', 'universi.structure']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.96.1',
 'type-inspector>=1.2.2,<2.0.0',
 'typer[all]',
 'typing-extensions']

setup_kwargs = {
    'name': 'universi',
    'version': '0.2.0',
    'description': '',
    'long_description': '# universi\n\nPackage description\n\n---\n\n<p align="center">\n<a href="https://github.com/ovsyanka83/universi/actions?query=workflow%3ATests+event%3Apush+branch%3Amain" target="_blank">\n    <img src="https://github.com/Ovsyanka83/universi/actions/workflows/test.yaml/badge.svg?branch=main&event=push" alt="Test">\n</a>\n<a href="https://codecov.io/gh/ovsyanka83/universi" target="_blank">\n    <img src="https://img.shields.io/codecov/c/github/ovsyanka83/universi?color=%2334D058" alt="Coverage">\n</a>\n<a href="https://pypi.org/project/universi/" target="_blank">\n    <img alt="PyPI" src="https://img.shields.io/pypi/v/universi?color=%2334D058&label=pypi%20package" alt="Package version">\n</a>\n<a href="https://pypi.org/project/universi/" target="_blank">\n    <img src="https://img.shields.io/pypi/pyversions/universi?color=%2334D058" alt="Supported Python versions">\n</a>\n</p>\n\n## Installation\n\n```bash\npip install universi\n```\n',
    'author': 'Stanislav Zmiev',
    'author_email': 'szmiev2000@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ovsyanka83/universi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
